# -*- coding: utf-8 -*-
#
# Copyright 2020 Osia authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module implements configuration object for aws installation"""
from typing import List, Optional
import logging
import boto3

from .base import AbstractInstaller


class AWSInstaller(AbstractInstaller):
    """Object containing all configuration related
    to aws installation"""
    def __init__(self,
                 cluster_region=None,
                 list_of_regions=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.cluster_region = cluster_region
        self.list_of_regions = list_of_regions if list_of_regions else []

    def get_template_name(self):
        return 'aws.jinja2'

    def acquire_resources(self):
        region = get_free_region(self.list_of_regions)

        if region is None:
            logging.error("No free region amongst selected ones: %s",
                          ', '.join(self.list_of_regions))
            raise Exception("No free region found")
        logging.info("Selected region %s", region)
        self.cluster_region = region

    def get_api_ip(self) -> Optional[str]:
        return None

    def get_apps_ip(self):
        return None

    def post_installation(self):
        pass


def get_free_region(order_list: List[str]) -> Optional[str]:
    """Finds first free region in provided list,
    if provided list is empty, it searches all regions"""
    candidates = order_list[:]
    if len(candidates) == 0:
        candidates = [v['RegionName'] for v in boto3.client('ec2').describe_regions()['Regions']]
    for candidate in candidates:
        region = boto3.client('ec2', candidate)
        count = len(region.describe_vpcs()['Vpcs'])
        if count < 5:
            logging.debug("Selected region %s", candidate)
            return candidate
    return None



# -*- coding: utf-8 -*-
#
# Copyright 2020 Osia authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module implements methods specific for nsupate provider"""
import logging
from subprocess import Popen, PIPE
from osia.installer.dns.base import DNSUtil
from osia.installer.clouds.base import AbstractInstaller


class NSUpdate(DNSUtil):
    """Implementation of DNSUtil specific for nsupdate dns provider"""
    def __init__(self, key_file=None, server=None, zone=None, use_ipv4=False, **kwargs):
        super().__init__(**kwargs)
        self.key_file = key_file
        self.server = server
        self.zone = zone
        self.use_ipv4 = use_ipv4

    def _exec_nsupdate(self, string: str):
        nsupdate_args = ["nsupdate", "-k", self.key_file]
        if self.use_ipv4:
            nsupdate_args.append('-4')
        with Popen(nsupdate_args, stdin=PIPE, universal_newlines=True) as nsu:
            nsu.communicate(string)
            self.modified = True

    def provider_name(self):
        return 'nsupdate'

    def _get_start(self):
        result = str()
        if self.server:
            result += f"server {self.server}\n"
        if self.zone:
            result += f"zone {self.zone}\n"
        return result

    def _get_suffix(self):
        return f"{self.cluster_name}.{self.base_domain}"

    def add_api_domain(self, instance: AbstractInstaller):
        if instance.get_api_ip() is None:
            logging.debug("Not applying dns settings since no ip address is associated")
            return
        logging.info("Adding api domain api.%s for floating ip addr %s",
                     self._get_suffix(), instance.get_api_ip())
        nsupdate_string = f"""{self._get_start()}
update add api.{self._get_suffix()} {self.ttl} A {instance.get_api_ip()}
send
"""
        self._exec_nsupdate(nsupdate_string)

    def add_apps_domain(self, instance: AbstractInstaller):
        if instance.get_apps_ip() is None:
            logging.debug("Not applying dns settings since no ip address is associated")
            return
        logging.info("Adding apps domain *.apps.%s for floating ip addr %s",
                     self._get_suffix(), instance.get_apps_ip())
        nsupdate_string = f"""{self._get_start()}
update add apps.{self._get_suffix()} {self.ttl} A {instance.get_apps_ip()}
update add \\*.apps.{self._get_suffix()} {self.ttl} A {instance.get_apps_ip()}
send
"""
        self._exec_nsupdate(nsupdate_string)

    def delete_domains(self):
        nsupdate_string = f"""{self._get_start()}
update delete apps.{self._get_suffix()} A
update delete \\*.apps.{self._get_suffix()} A
update delete api.{self._get_suffix()} A
send"""

        self._exec_nsupdate(nsupdate_string)
        self.delete_file()



# -*- coding: utf-8 -*-
#
# Copyright 2020 Osia authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module implements common base for install-config
creation.
It also implements logic to obtain correct specification
for specified installation platform"""

import logging

from abc import abstractmethod, ABC
from subprocess import run
from typing import ClassVar, Optional
from jinja2 import Environment, PackageLoader
from semantic_version import Version, SimpleSpec


class AbstractInstaller(ABC):
    """Base object for configuration of install-config"""
    # pylint: disable=too-many-instance-attributes

    __env: Environment = None

    # pylint: disable=too-many-arguments
    def __init__(self,
                 cluster_name=None,
                 base_domain=None,
                 master_flavor=None,
                 master_replicas='3',
                 pull_secret_file=None,
                 ssh_key_file=None,
                 worker_flavor=None,
                 worker_replicas='3',
                 certificate_bundle_file=None,
                 cluster_directory=None,
                 skip_clean=False,
                 enable_fips=False,
                 installer=None,
                 **unused_kwargs):
        self.cluster_name = cluster_name
        self.base_domain = base_domain
        self.master_flavor = master_flavor
        self.master_replicas = master_replicas
        self.pull_secret_file = pull_secret_file
        self.pull_secret = None
        self.ssh_key_file = ssh_key_file
        self.ssh_key = None
        self.certificate_bundle_file = certificate_bundle_file
        self.certificate_bundle = None
        self.worker_flavor = worker_flavor
        self.worker_replicas = worker_replicas
        self.cluster_directory = cluster_directory
        self.skip_clean = skip_clean
        self.installer = installer
        self.enable_fips = enable_fips
        self.ocp_version = None
        self.network_type = "OVNKubernetes"

    @abstractmethod
    def acquire_resources(self):
        """Method which is called before the install-config can be generated.
        The method should be used to get all necessary dependencies to fill
        in details in install-config"""

    @abstractmethod
    def get_template_name(self):
        """Method to obtain name of jinja template related to specified platform."""

    @abstractmethod
    def post_installation(self):
        """Method called after the installation is done, this is the place where,
        things like registration of apps domain and other is expected to happen."""

    @abstractmethod
    def get_api_ip(self) -> Optional[str]:
        """Returns api ip address if dns is supported, None otherwise.
        """

    @abstractmethod
    def get_apps_ip(self):
        """Returns apps ip if dns is supported, None otherwise
        """

    def _resolve_version(self):
        if self.ocp_version is None:
            capture = run([self.installer, "version"], capture_output=True, check=False)
            ver_string = capture.stdout.decode("utf-8").splitlines()[0].split(' ')[1]
            self.ocp_version = ver_string
            logging.info("Resolved installed version as %s", self.ocp_version)

    def _resolve_network_type(self):
        self._resolve_version()
        ver = Version(self.ocp_version)
        spec = SimpleSpec('<4.15')
        if ver in spec:
            logging.debug("Older version of openshift, falling back to OpenShiftSDN")
            self.network_type = 'OpenShiftSDN'

    def check_clean(self):
        """Method returns if installation is configured to clean resources
        on failure."""
        return not self.skip_clean

    def process_template(self):
        """Method executes creation of install-config.yaml"""
        self._resolve_network_type()
        with open(self.pull_secret_file) as ps_file:
            self.pull_secret = ps_file.read()
        with open(self.ssh_key_file) as key_file:
            self.ssh_key = key_file.read()
        if self.certificate_bundle_file is not None:
            with open(self.certificate_bundle_file) as cert_file:
                self.certificate_bundle = cert_file.read()
        template = AbstractInstaller.get_environment().get_template(self.get_template_name())
        result = template.render(self.__dict__)
        with open(f"{self.cluster_name}/install-config.yaml", "w") as yaml_file:
            yaml_file.write(result)

    @classmethod
    def get_environment(cls) -> Environment:
        """Method loads jinja templates"""
        if cls.__env is None:
            cls.__env = Environment(loader=PackageLoader("osia.installer"))
        return cls.__env


class InstallerProvider:
    """Class implements dynamic provider of registered platform specific
    implementations of AbstractInstaller.
    Class implements singleton design pattern."""
    __instance: "InstallerProvider" = None

    @classmethod
    def instance(cls) -> "InstallerProvider":
        """Method to obtains singleton instance."""
        if cls.__instance is None:
            cls.__instance = cls()
        return cls.__instance

    @classmethod
    def register(cls, name, instance):
        """Method to dynamically register implementation of AbstractInstaller"""
        cls.instance().add_installer(name, instance)

    def __init__(self):
        self.installers = {}

    def add_installer(self, name: str, instance: ClassVar):
        """Insert concrete implementation of AbstractInstaller into the registry"""
        self.installers[name] = instance

    def __getitem__(self, name: str) -> ClassVar:
        return self.installers[name]



"""Module implements logic for rhcos image download"""
import subprocess
from subprocess import Popen, PIPE
from shutil import copyfileobj
from pathlib import Path
from typing import Tuple
from tempfile import NamedTemporaryFile


import gzip
import re
import logging
import json

import requests
from .utils import get_data

GITHUB_URL = "https://raw.githubusercontent.com/openshift/installer/{commit}/data/data/rhcos.json"


class CoreOsException(Exception):
    """CoreOsException represents error while executing installer in older version
    """
    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, *kwargs)


def _get_coreos_json(installer: str) -> str:
    json_data = {}
    with Popen([installer, "coreos", "print-stream-json"], stdout=PIPE,
               stderr=subprocess.DEVNULL, universal_newlines=True) as proc:
        proc.wait()
        if proc.returncode != 0:
            raise CoreOsException("Installer doesn't support coreos subcommand")
        json_data = json.loads(proc.stdout.read())
    json_part = json_data["architectures"]["x86_64"]["artifacts"]["openstack"]
    release_str = json_part["release"]
    json_part = json_part["formats"]["qcow2.gz"]
    return json_part.get("disk", json_part)["location"], release_str


def get_commit(installer: str) -> Tuple[str, str]:
    """Function extracts source commit from installer,
    in order to find associated rhcos image"""
    version_str = ""
    commit_regex = re.compile(r"^.*commit (?P<commit>\w*)$", re.MULTILINE)
    with Popen([installer, "version"], stdout=PIPE, universal_newlines=True) as proc:
        version_str = proc.stdout.read()
    commits = commit_regex.findall(version_str)
    logging.info("Found commits by running installer %s", commits)
    return commits[0]


def _get_old_url(installer: str) -> Tuple[str, str]:
    commit = get_commit(installer)
    gh_data_link = GITHUB_URL.format(commit=commit)
    rhcos_json = requests.get(gh_data_link, allow_redirects=True)
    rhcos_data = json.loads(rhcos_json.content)
    return rhcos_data['baseURI'] + rhcos_data['images']['openstack']['path'], rhcos_data['buildid']


def get_url(installer: str) -> Tuple[str, str]:
    """Function builds url to rhcos image and version of
    rhcos iamge."""
    url, version = None, None
    try:
        url, version = _get_coreos_json(installer)
    except CoreOsException as ex:
        logging.debug(ex)
        url, version = _get_old_url(installer)
    return url, version


def _extract_gzip(buff: NamedTemporaryFile, target: str) -> Path:
    result = None
    with gzip.open(buff.name) as zip_file:
        result = Path(target)
        with result.open("wb") as output:
            copyfileobj(zip_file, output)
    return result


def download_image(image_url: str, image_file: str):
    """Main entrypoint for image download, function
    extracts url to rhcos image, downloads and extracts it
    to specified target"""
    directory = Path(image_file).parent
    # Check if the directory exists
    if not directory.exists():
        # If the directory does not exist, create it
        directory.mkdir(parents=True)
        logging.debug("Creating %s directory for download images", directory)
    else:
        logging.debug("Directory %s for images already exists", directory)
    res_file = get_data(image_url, image_file, _extract_gzip)
    return res_file



# -*- coding: utf-8 -*-
#
# Copyright 2020 Osia authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module exposes configuration to setup DNS needed for Openshift to work"""
from osia.installer.dns.nsupdate import NSUpdate
from osia.installer.dns.route53 import Route53Provider
from osia.installer.dns.base import DNSProvider

DNSProvider.register_provider('nsupdate', NSUpdate)
DNSProvider.register_provider('route53', Route53Provider)



# -*- coding: utf-8 -*-
#
# Copyright 2020 Osia authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module responsible for download of openshift-install binary"""
from shutil import copyfileobj
from tempfile import NamedTemporaryFile
from pathlib import Path
from typing import Tuple, Optional
import platform

import logging
import re
import stat
import tarfile
import time
import requests

from bs4 import BeautifulSoup
from .utils import get_data


PROD_ROOT = "http://mirror.openshift.com/pub/openshift-v4/{}/clients/ocp/"
BUILD_ROOT = "https://openshift-release-artifacts.svc.ci.openshift.org/"
PREVIEW_ROOT = "http://mirror.openshift.com/pub/openshift-v4/{}/clients/ocp-dev-preview/"

VERSION_RE = re.compile(r"^openshift-install(-rhel(?P<rhel>\d+))?(-(?P<platform>(linux|mac)))?"
                        r"(-(?P<architecture>\w+))?(-(?P<version>\d+.*))?\.tar\.gz")
EXTRACTION_RE = re.compile(r'.*Extracting tools for .*, may take up to a minute.*')


def _current_platform() -> Tuple[str, str]:
    if platform.system() == "Linux" and platform.machine() == "x86_64":
        return "linux", "amd64"
    if platform.system() == "Linux" and (
            platform.machine() == "arm64" or platform.machine() == "aarch64"):
        return "linux", "arm64"
    if platform.system() == "Darwin" and platform.machine() == "arm64":
        return "mac", "arm64"
    if platform.system() == "Darwin" and platform.machine() == "x86_64":
        return "mac", "amd64"

    raise Exception(f"Unrecognized platform {platform.system()} {platform.machine()}")


def get_url(directory: str, arch: str, fips: bool = False,
            rhel_version: str = None) -> Tuple[Optional[str], Optional[str]]:
    """Searches the http directory and returns both url to installer
    and version.
    """
    if fips and not rhel_version:
        raise Exception("Rhel version was not detected. Please download installer separatly.")

    logging.debug('Url for installers look-up %s', directory)
    lst = requests.get(directory, allow_redirects=True)
    tree = BeautifulSoup(lst.content, 'html.parser')
    links = tree.find_all('a')
    installer, version = None, None
    os_name, local_arch = _current_platform()
    for k in links:
        logging.debug('Parsing link: %s', k.get('href'))
        match = VERSION_RE.match(k.get('href'))

        if match:
            if match.group("version"):
                version = match.group('version')

            if fips and match.group("rhel") == rhel_version:
                installer = lst.url + k.get('href')
                break

            if not fips and match.group('platform') == os_name:
                if (local_arch == match.group('architecture')) \
                        or (local_arch == arch and not match.group('architecture')):
                    installer = lst.url + k.get('href')
                    break
    else:
        if fips:
            raise Exception(f"FIPS Installer not found for {rhel_version=}")
        raise Exception("Installer not found")
    return installer, version


def get_devel_url(version: str, arch: str, fips: bool = False,
                  rhel_version: str = None) -> Tuple[Optional[str], Optional[str]]:
    """
    Searches developement sources and returns url to installer
    """
    req = requests.get(BUILD_ROOT + version + "/", allow_redirects=True)
    ast = BeautifulSoup(req.content, 'html.parser')
    logging.info('Checking stage repository for installer')
    while len(ast.find_all('p')) != 0 and EXTRACTION_RE.match(next(ast.find_all('p')[0].children)):
        logging.debug('The installer was not extracted yet, waiting for 10s')
        time.sleep(10)
        req = requests.get(BUILD_ROOT + version, allow_redirects=True)
        ast = BeautifulSoup(req.content, 'html.parser')
    logging.debug('Installer found on page, continuing')
    return get_url(req.url, arch, fips, rhel_version)


def get_prev_url(version: str, arch: str, fips: bool = False,
                 rhel_version: str = None) -> Tuple[Optional[str], Optional[str]]:
    """Returns installer url from dev-preview sources"""
    return get_url(PREVIEW_ROOT.format(arch) + version + "/", arch, fips, rhel_version)


def get_prod_url(version: str, arch: str, fips: bool = False,
                 rhel_version: str = None) -> Tuple[Optional[str], Optional[str]]:
    """Returns installer url from production sources"""
    return get_url(PROD_ROOT.format(arch) + version + "/", arch, fips, rhel_version)


def _get_storage_path(version: str, install_base: str) -> str:
    path = Path(install_base)
    spec_path = path.joinpath(version)
    if not spec_path.exists():
        spec_path.mkdir()
    return spec_path.as_posix()


def _extract_tar(buffer: NamedTemporaryFile, target: str) -> Path:
    result = None
    with tarfile.open(buffer.name) as tar:
        inst_info = None
        for i in tar.getmembers():
            if i.name in ['openshift-install', 'openshift-install-fips']:
                inst_info = i
        if inst_info is None:
            raise Exception("error")
        stream = tar.extractfile(inst_info)
        result = Path(target).joinpath(inst_info.name)
        with result.open('wb') as output:
            copyfileobj(stream, output)
        result.chmod(result.stat().st_mode | stat.S_IXUSR)
    return result


def get_installer(tar_url: str, target: str):
    """Download and extract the installer into the target"""
    return get_data(tar_url, target, _extract_tar)


# pylint: disable=too-many-arguments
def download_installer(installer_version: str,
                       installer_arch: str,
                       dest_directory: str,
                       source: str,
                       fips: bool = False,
                       rhel_version: str = None) -> str:
    """Starts search and extraction of installer"""
    logging.debug("Getting version %s of %s, storing to directory %s and devel is %r",
                  installer_version, installer_arch, dest_directory, source)

    downloader = None
    if source == "prod":
        downloader = get_prod_url
    elif source == "prev":
        downloader = get_prev_url
    elif source == "devel":
        downloader = get_devel_url
    else:
        raise Exception("Error for source profile " + source)

    url, version = downloader(installer_version, installer_arch, fips, rhel_version)
    logging.debug('Installer\'s URL is  %s and full version is %s', url, version)
    root = Path(dest_directory).joinpath(version)

    installer_exe_name = 'openshift-install'

    if fips:
        installer_exe_name = 'openshift-install-fips'

    if root.exists() and root.joinpath(installer_exe_name).exists():
        logging.info('Found installer at %s', root.as_posix())
        return root.joinpath(installer_exe_name).as_posix()
    root.mkdir(parents=True, exist_ok=True)
    return get_installer(url, root.as_posix())



# -*- coding: utf-8 -*-
#
# Copyright 2020 Osia authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module implements support for Openstack installation"""
from operator import itemgetter
from pathlib import Path
from typing import List, Optional, Tuple
from os import path

import json
import warnings
import logging

from openstack.connection import from_config, Connection
from openstack.network.v2.floating_ip import FloatingIP
from openstack.network.v2.port import Port
from openstack.image.v2.image import Image
from openstack.exceptions import SDKException
from osia.installer.clouds.base import AbstractInstaller
from osia.installer.downloader import get_url, download_image


class ImageException(Exception):
    """
    Image exception encapsulates Exception happening while image resolution
    """
    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)


def _load_connection_openstack(conn_name: str, args=None) -> Connection:
    connection = from_config(cloud=conn_name, options=args)
    if connection is None:
        raise Exception(f"Unable to connect to ${conn_name}")
    return connection


def _update_json(json_file: str, fip: str):
    res = None
    with open(json_file) as inp:
        res = json.load(inp)

    res['fips'].append(fip)
    with open(json_file, "w") as out:
        json.dump(res, out)


def delete_fips(fips_file: str):
    """Deletes floating ips stored in configuration file"""
    fips = None
    with open(fips_file) as fi_file:
        fips = json.load(fi_file)
    connection = _load_connection_openstack(fips['cloud'])
    os_fips = [j for k in fips['fips'] for j in connection.network.ips(floating_ip_address=k)]
    for i in os_fips:
        connection.network.delete_ip(i)


def delete_image(fips_file, cluster_name):
    """Function checkes the uploaded image into openstack
    and removes from its metadata information about associated
    cluster.
    If last associated cluster was removed, the image is deleted"""
    fips = None
    with open(fips_file) as json_file:
        fips = json.load(json_file)
    if fips.get('image', None) is None:
        return
    connection = _load_connection_openstack(fips['cloud'])
    image = connection.image.find_image(fips['image'])
    clusters = image.properties['osia_clusters'].split(',')
    clusters.remove(cluster_name)
    if len(clusters) == 0:
        logging.info("Deleting uploaded image %s, since all clusters were removed", image.name)
        connection.image.delete_image(image)
    else:
        logging.info("Removing cluster %s from image %s metadata", cluster_name, image.name)
        connection.image.update_image(image, osia_clusters=','.join(clusters))


def _find_best_fit(networks: dict) -> str:
    return max(networks.items(), key=itemgetter(1))[0]


def _find_fit_network(osp_connection: Connection,
                      networks: List[str]) -> Tuple[Optional[str], Optional[str]]:
    named_networks = {k['name']: k for k in osp_connection.list_networks() if k['name'] in networks}
    results = {}
    for net_name in networks:
        net_avail = osp_connection.network.get_network_ip_availability(named_networks[net_name].id)
        subnet_usage = [(subnet['total_ips'], subnet['used_ips'])
                        for subnet in net_avail.subnet_ip_availability if subnet['ip_version'] == 4]
        total_ips, used_ips = [sum(i) for i in zip(*subnet_usage)]
        results[net_name] = total_ips / used_ips
    result = _find_best_fit(results)
    return named_networks[result]['id'], result


def _find_cluster_ports(osp_connection: Connection, cluster_name: str) -> Port:
    port_list = [k for k in osp_connection.list_ports()
                 if k.name.startswith(cluster_name) and k.name.endswith('ingress-port')]
    port = next(iter(port_list), None)
    if port is None:
        raise Exception(f"Ingress port for cluster {cluster_name} was not found")
    return port


def _attach_fip_to_port(osp_connection: Connection, fip_addr, ingress_port):
    osp_connection.network.add_ip_to_port(ingress_port, fip_addr)


def _get_floating_ip(osp_connection: Connection,
                     cloud: str,
                     network_id: str,
                     cluster_name: str,
                     purpose: str) -> FloatingIP:
    fip = osp_connection.network.create_ip(floating_network_id=network_id,
                                           description=f"{cluster_name}-{purpose}")
    if fip is None:
        raise Exception(f"Allocation of Ip failed for network ${network_id}")

    if path.exists(f"{cluster_name}/fips.json"):
        _update_json(f"{cluster_name}/fips.json", fip.floating_ip_address)
    else:
        with open(f"{cluster_name}/fips.json", "w") as fips:
            conf = {'cloud': cloud, 'fips': [fip.floating_ip_address]}
            json.dump(conf, fips)
    return fip


def add_cluster(osp_connection: Connection, image: Image, cluster_name: str):
    """Function adds cluster name to image metadata in order to prevent
    image deletion"""
    clusters = image.properties['osia_clusters'].split(',')
    clusters.append(cluster_name)
    osp_connection.image.update_image(image, osia_clusters=','.join(clusters))


def upload_uniq_image(osp_connection: Connection,
                      cloud: str,
                      cluster_name: str,
                      images_dir: str,
                      installer: str):
    """Function uploads unique image to the cluster, instead of making shared one"""
    inst_url, version = get_url(installer)
    image_name = f"osia-{cluster_name}-{version}"
    image_path = Path(images_dir).joinpath(f"rhcos-{version}.qcow2")
    image_file = None
    if image_path.exists():
        logging.info("Found image at %s", image_path.name)
        image_file = image_path.as_posix()
    else:
        logging.info("Starting download of image %s", inst_url)
        image_file = download_image(inst_url, image_path.as_posix())

    logging.info("Starting upload of image into openstack")
    osp_connection.create_image(image_name, filename=image_file,
                                container_format="bare", disk_format="qcow2", wait=True,
                                osia_clusters=cluster_name, visibility='private')
    logging.info("Upload finished")
    image = osp_connection.image.find_image(image_name)
    logging.info("Image uploaded as %s", image.name)
    with open(Path(cluster_name).joinpath("fips.json"), "w") as fips:
        obj = {'cloud': cloud, 'fips': [], 'image': image_name}
        json.dump(obj, fips)
    return image.name


# pylint: disable=too-many-arguments
def resolve_image(osp_connection: Connection,
                  cloud: str,
                  cluster_name: str,
                  images_dir: str,
                  installer: str,
                  error: Optional[Exception]):
    """Function searches for image in openstack and creates it
    if it doesn't exist"""
    inst_url, version = get_url(installer)
    image_name = f"osia-rhcos-{version}"
    image = osp_connection.image.find_image(image_name, ignore_missing=True)
    if image is None:
        image_path = Path(images_dir).joinpath(f"rhcos-{version}.qcow2")
        image_file = None
        if image_path.exists():
            logging.info("Found image at %s", image_path.name)
            image_file = image_path.as_posix()
        else:
            logging.info("Starting download of image %s", inst_url)
            image_file = download_image(inst_url, image_path.as_posix())

        logging.info("Starting upload of image into openstack")
        osp_connection.create_image(image_name, filename=image_file,
                                    container_format="bare", disk_format="qcow2", wait=True,
                                    osia_clusters=cluster_name, visibility='private')
        logging.info("Upload finished")
        image = osp_connection.image.find_image(image_name)
        logging.info("Image uploaded as %s", image.name)
    else:
        logging.info("Reusing found image in openstack %s", image.name)
        try:
            add_cluster(osp_connection, image, cluster_name)
        except SDKException as err:
            if error is not None:
                raise ImageException("Couldn't add cluster to image") from err
            logging.warning("Image disappeared while metadata were written, trying again")
            logging.debug("Openstack error: %s", err)
            return resolve_image(osp_connection, cloud, cluster_name, images_dir, installer, err)
    with open(Path(cluster_name).joinpath("fips.json"), "w") as fips:
        obj = {'cloud': cloud, 'fips': [], 'image': image_name}
        json.dump(obj, fips)
    return image.name


class OpenstackInstaller(AbstractInstaller):
    """Class containing configuration related to openstack"""
    # pylint: disable=too-many-instance-attributes,too-many-arguments
    def __init__(self,
                 osp_cloud=None,
                 osp_base_flavor=None,
                 network_list=None,
                 os_image=None,
                 images_dir=None,
                 osp_image_download=False,
                 osp_image_unique=False,
                 args=None,
                 **kwargs):
        super().__init__(**kwargs)
        if 'psi_base_flavor' in kwargs:
            warnings.warn("[Deprecation warning]: option psi_base_flavor was replaced by option "
                          "osp_base_flavor this option will be removed in next releases.")
            self.osp_base_flavor = kwargs['psi_base_flavor']
        if 'psi_cloud' in kwargs:
            warnings.warn("[Deprecation warning]: option psi_cloud was replaced by option osp_cloud"
                          "this option will be removed in next releases.")
            self.osp_cloud = kwargs['psi_cloud']

        if osp_cloud is not None:
            self.osp_cloud = osp_cloud
        if osp_base_flavor is not None:
            self.osp_base_flavor = osp_base_flavor
        self.network_list = network_list
        self.args = args
        self.os_image = os_image
        self.image_download = osp_image_download
        self.image_uniq = osp_image_unique
        self.osp_fip = None
        self.network = None
        self.connection = None
        self.apps_fip = None
        self.osp_network = None
        self.images_dir = images_dir

    def get_template_name(self):
        return 'openstack.jinja2'

    def acquire_resources(self):
        self.connection = _load_connection_openstack(self.osp_cloud)
        if self.image_uniq and (self.os_image is None or self.os_image == ""):
            self.os_image = upload_uniq_image(self.connection, self.osp_cloud, self.cluster_name,
                                              self.images_dir, self.installer)
        elif self.image_download and (self.os_image is None or self.os_image == ""):
            self.os_image = resolve_image(self.connection, self.osp_cloud, self.cluster_name,
                                          self.images_dir, self.installer, None)
        self.network, self.osp_network = _find_fit_network(self.connection, self.network_list)
        if self.network is None:
            raise Exception("No suitable network found")
        self.osp_fip = _get_floating_ip(self.connection,
                                        self.osp_cloud,
                                        self.network,
                                        self.cluster_name,
                                        "api").floating_ip_address

    def post_installation(self):
        ingress_port = _find_cluster_ports(self.connection, self.cluster_name)
        apps_fip = _get_floating_ip(self.connection,
                                    self.osp_cloud,
                                    self.network,
                                    self.cluster_name,
                                    "ingress")
        _attach_fip_to_port(self.connection, apps_fip, ingress_port)
        self.apps_fip = apps_fip.floating_ip_address

    def get_api_ip(self) -> Optional[str]:
        return self.osp_fip

    def get_apps_ip(self):
        return self.apps_fip
